# -*- coding: UTF-8 -*-
# 一个简单的finetuning小例子，先在0-7的8类上训练一个分类模型
# 然后使用这个模型在0-9的10类问题上做finetuning

import os
import numpy as np
import tensorflow as tf
from tensorflow.contrib.learn.python.learn.datasets.mnist import read_data_sets, DataSet


def build_8class_model(image, label):
    with tf.variable_scope('conv'):
        conv1 = tf.layers.conv2d(image, 16, 3, strides=1, padding='same',
                                 kernel_initializer=tf.truncated_normal_initializer(stddev=0.1),
                                 name='conv1')
        act1 = tf.nn.relu(conv1, name='act1')
        pool1 = tf.layers.max_pooling2d(act1, [2, 2], strides=[2, 2], padding='same',
                                        name='pool1')
        flatten_layer = tf.contrib.layers.flatten(pool1, 'flatten_layer')
    weights = tf.get_variable(shape=[flatten_layer.shape[-1], 8], dtype=tf.float32,
                              initializer=tf.truncated_normal_initializer(stddev=0.1), name='weights8')
    biases = tf.get_variable(shape=[8], dtype=tf.float32,
                             initializer=tf.constant_initializer(0.0), name='biases8')
    logit_output = tf.nn.bias_add(tf.matmul(flatten_layer, weights), biases, name='logit_output8')
    one_hot_label = tf.one_hot(indices=label, depth=8)
    cross_entropy = tf.reduce_mean(
        tf.nn.softmax_cross_entropy_with_logits(labels=one_hot_label, logits=logit_output))
    train_step = tf.train.GradientDescentOptimizer(0.1).minimize(cross_entropy)
    correct_prediction = tf.equal(tf.argmax(logit_output, 1), label)
    accuracy = tf.reduce_mean(tf.cast(correct_prediction, tf.float32))

    return train_step, cross_entropy, accuracy


def build_10class_model(image, label):
    with tf.variable_scope('conv'):
        conv1 = tf.layers.conv2d(image, 16, 3, strides=1, padding='same',
                                 kernel_initializer=tf.truncated_normal_initializer(stddev=0.1),
                                 name='conv1')
        act1 = tf.nn.relu(conv1, name='act1')
        pool1 = tf.layers.max_pooling2d(act1, [2, 2], strides=[2, 2], padding='same',
                                        name='pool1')
        flatten_layer = tf.contrib.layers.flatten(pool1, 'flatten_layer')
    weights = tf.get_variable(shape=[flatten_layer.shape[-1], 10], dtype=tf.float32,
                              initializer=tf.truncated_normal_initializer(stddev=0.1), name='weights10')
    biases = tf.get_variable(shape=[10], dtype=tf.float32,
                             initializer=tf.constant_initializer(0.0), name='biases10')
    logit_output = tf.nn.bias_add(tf.matmul(flatten_layer, weights), biases, name='logit_output10')
    one_hot_label = tf.one_hot(indices=label, depth=10)
    cross_entropy = tf.reduce_mean(
        tf.nn.softmax_cross_entropy_with_logits(labels=one_hot_label, logits=logit_output))
    train_step = tf.train.GradientDescentOptimizer(0.1).minimize(cross_entropy)
    correct_prediction = tf.equal(tf.argmax(logit_output, 1), label)
    accuracy = tf.reduce_mean(tf.cast(correct_prediction, tf.float32))
    return train_step, cross_entropy, accuracy


if __name__ == "__main__":
    folder_path = 'MNIST_data'  # your folder to put mnist data
    # prepare data for 8&10 class classification
    mnist_data = read_data_sets(folder_path, one_hot=False, reshape=False)
    testlabels_idx8 = np.where(mnist_data.test.labels < 8)
    trainlabels_idx8 = np.where(mnist_data.train.labels < 8)
    validationlabels_idx8 = np.where(mnist_data.validation.labels < 8)
    test_images8 = mnist_data.test.images[testlabels_idx8]
    test_labels8 = mnist_data.test.labels[testlabels_idx8]
    train_images8 = mnist_data.train.images[trainlabels_idx8]
    train_labels8 = mnist_data.train.labels[trainlabels_idx8]
    validation_images8 = mnist_data.validation.images[validationlabels_idx8]
    validation_labels8 = mnist_data.validation.labels[validationlabels_idx8]

    options = dict(dtype=tf.uint8, reshape=False, seed=None)
    train8 = DataSet(train_images8, train_labels8, **options)
    validation8 = DataSet(validation_images8, validation_labels8, **options)

    # construct network
    image = tf.placeholder(dtype=tf.float32, shape=[None, 28, 28, 1], name='image')
    label = tf.placeholder(tf.int64, [None])
    train_step, cross_entropy, accuracy = build_8class_model(image, label)  # for 8 class
    # train_step, cross_entropy, accuracy = build_10class_model(image, label) # for 10 class

    # initialize network
    sess = tf.InteractiveSession()
    tf.global_variables_initializer().run()
    if tf.train.latest_checkpoint('ckpts') is not None:
        trainable_vars = tf.trainable_variables()
        res_vars = [t for t in trainable_vars if t.name.startswith('conv')]
        saver = tf.train.Saver(var_list=res_vars)
        saver.restore(sess, tf.train.latest_checkpoint('ckpts'))
    else:
        saver = tf.train.Saver()

    # start to train
    for i in range(2001):
        # for 8 class
        batch_xs, batch_ys = train8.next_batch(100)
        # for 10 class
        # batch_xs, batch_ys = mnist_data.train.next_batch(100)

        _ = sess.run(train_step, feed_dict={image: batch_xs, label: batch_ys})
        if i % 100 == 0:
            # for 8 class
            loss_res, acc_res = sess.run([cross_entropy, accuracy],
                                         feed_dict={image: validation8.images, label: validation8.labels})
            # for 10 class
            # loss_res, acc_res = sess.run([cross_entropy, accuracy],
            #                              feed_dict={image: mnist_data.validation.images,
            #                                         label: mnist_data.validation.labels})
            print 'step:%d,loss:%f,acc:%f' % (i, loss_res, acc_res)
        if i % 1000 == 0:
            # for 8 class
            save_path = os.path.join('8class-ckpts', str(i) + '-model.ckpt')
            # for 10 class
            save_path = os.path.join('10class-ckpts', 'ft_' + str(i) + '-model.ckpt')
            saver.save(sess, save_path)

    print 'train finished'
